/* $Workfile: GimleGroomer.c $       $Revision: 1 $      $Modtime: 4/03/12 12:53p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    GimleGroomer.c

    A simple example console application demonstrating how to create an E1/T1
    grooming application using a number of Gimle-16-PCI-Plus (or Gimle-16-PCIe) 
    boards connected to a number of Thor boards (Thor-2-PCI-Plus, Thor-4-PCI-Plus, 
    Thor-8-PCI-Plus) over the H.100 bus.

    The Gimle board receives a number of timeslots from the 16 E1/T1 spans. The
    timeslots are cross-connected to selected timeslots on outgoing E1/T1 spans
    on the Thor board.

    The cross connections are defined in a separate text file (GimleGroomer.con)
    in the following format:

    # Gimle Brd#, Gimle Li, Gimle TS   -> Thor Brd#, Thor Li,   Thor TS

    # Example:
       0            0         31             0            0         1
       0            1         16             0            0         2
       0           15          5             0            1         1
       0           16         16             0            1         2

    The T1/E1 configuration parameters are supplied by command line parameters.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include <math.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_CON_FILE "GimleGroomer.con"
#define DEMO_DEFAULT_LOG_FILE "GimleGroomer.log"
#define DEMO_DEFAULT_GIMLE_BOARD_NO 0
#define DEMO_DEFAULT_THOR_BOARD_NO  0
#define DEMO_CO_GIMLE_LI 16
#define DEMO_MAX_BOARDS 32

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  OTX_BOOL               bEnumerateOnly;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  char                   conFileName[OTX_MAX_PATH];

  OtxT1E1LiModeE         eGimleT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bGimleReceiveEqualizer;
  OtxT1E1TermModeE       eGimleTermMode;
  OTX_BOOL               bGimleMonitorMode;
  OTX_BOOL               bGimleAmplifier;
  OTX_INT32              nGimleRxThreshold;
  OtxT1LineCodeE         eGimleT1LineCode;
  OtxT1FrameFormatE      eGimleT1FrameFormat;
  OtxE1LineCodeE         eGimleE1LineCode;
  OtxE1FrameFormatE      eGimleE1FrameFormat;
  OtxBrdClockSourceE     eGimleClkSource;
  OTX_INT32              nGimleClkSourceBrd;  // Board number of the Gimle board that will be the clock source for the system

  OtxT1E1LiModeE         eThorT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bThorReceiveEqualizer;
  OtxT1E1TermModeE       eThorTermMode;
  OTX_BOOL               bThorMonitorMode;
  OTX_BOOL               bThorAmplifier;
  OTX_INT32              nThorRxThreshold;
  OtxT1LineCodeE         eThorT1LineCode;
  OtxT1FrameFormatE      eThorT1FrameFormat;
  OtxE1LineCodeE         eThorE1LineCode;
  OtxE1FrameFormatE      eThorE1FrameFormat;

} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE hDevice);
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice);
static signed int DemoGetNicDeviceNo(IN OTX_HANDLE hDevice);
static void DemoOpenThorBoard(OTX_INT32 eThorBoardType, OTX_INT32 nThorBoardNo);
static void DemoOpenGimleBoard(OTX_INT32 nGimleBoardNo);
static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_HANDLE  g_hCtBus;
OTX_INT32   g_coGimleBrds = 0;
OTX_INT32   g_coThorBrds = 0;

struct PhysDevsGimle {
  OTX_HANDLE hBoard;
  OTX_HANDLE hQuadE1T1[4]; 
  OTX_HANDLE hLi[16];
} g_hPhysDevsGimle[DEMO_MAX_BOARDS];

struct PhysDevsThor {
  OTX_HANDLE hBoard;
  OTX_HANDLE hQuadE1T1[4]; 
  OTX_HANDLE hLi[8];
  int        coThorLi;
} g_hPhysDevsThor[DEMO_MAX_BOARDS];


typedef struct DemoT1E1CfgDataT {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} DemoT1E1CfgDataT;

DemoT1E1CfgDataT g_ThorT1E1CfgData;
DemoT1E1CfgDataT g_GimleT1E1CfgData;



/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;   
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}



/*******************************************************************************

  @func

    Read board serial number

*******************************************************************************/
static void DemoGetBoardInformation(OTX_HANDLE hParentDevice, OTX_HANDLE hDevice, 
                          OTX_INT32 eDevType, OTX_UINT32 *nSubSystem, OTX_UINT32 *nBrdNo, char *pszSerNo)
{
  OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
	OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
	OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
	OTX_RESULT      result;
	char *pszSerialNumber;
  char *pszBrdBuildNumber;
  char *pszCpldBuildNumbers;
  OTX_UINT32 nSubsystemId;
  OTX_UINT32 nDeviceId;
  OTX_UINT32 nDeviceNo;
				
  if((hParentDevice == 0) && (eDevType != -1))		// Parent device with children
	{
		DemoLogf("\n");

    DemoLogf("%s Board\n", DemoGetDeviceTypeString(hDevice));

    // Read Device Number
    result = OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNo); DemoCheckRc(result);
    DemoLogf("Board Number       : %d\n", nDeviceNo);
    *nBrdNo = nDeviceNo;
    
    result = OtxDrvGetAttributeType(hDevice, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
		assert(eAttribType == OTX_ATTRIB_SZ);
		assert(bReadOnly == OTX_TRUE);
		pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
		result = OtxDrvGetAttributeValueSz(hDevice, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
		DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    strcpy(pszSerNo, pszSerialNumber);
    OsaThreadSleep(100); 
		free(pszSerialNumber);

    // Read Device ID (Board type)
    result = OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(result);
    DemoLogf("Board Device ID    : 0x%.4X\n", nDeviceId);


    // Read Subsystem Number (Product version) from the board
    result = OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_BRD_SUBSYSTEM_ID, &nSubsystemId); DemoCheckRc(result);
    DemoLogf("Board Subsystem ID : 0x%.4X\n", nSubsystemId);
    *nSubSystem = nSubsystemId;

    // Read Build Number from the board
    result = OtxDrvGetAttributeType(hDevice, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(hDevice, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number : %s\n", pszBrdBuildNumber);
    OsaThreadSleep(100); 
    free(pszBrdBuildNumber);

    if (eDevType != OTX_DEVICE_THOR_PCMCIA) {
      // Read CPLD Build Numbers from the board
      result = OtxDrvGetAttributeType(hDevice, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(hDevice, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
      DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
      OsaThreadSleep(100); 
      free(pszCpldBuildNumbers);
    }
		DemoLogf("\n");
  }
}



/*******************************************************************************

  @func

  Enumerates all devices for a given parent device 

  @comm

*******************************************************************************/
static void DemoEnumerateDevices()
{
  OTX_INT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;
  OTX_HANDLE hParentDevice = 0;

  DemoLogf("Entering EnumDevices...\n");

  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(hParentDevice, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(hParentDevice, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {

        if ((hParentDevice == 0) && (eDevType != (OTX_INT32) -1)) {		
          switch (eDevType) {
          case OTX_DEVICE_GIMLE_16_PCI_PLUS:
          case OTX_DEVICE_GIMLE_16_PCIE:
            //case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
            //case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
            {
              OTX_UINT32 nSubSystemId;
              OTX_UINT32 nBoardNo;
              char szSerNo[128];
              
              // Read serial number from the board
              DemoGetBoardInformation( hParentDevice, hChildDevice, eDevType, &nSubSystemId, &nBoardNo, szSerNo);

              if (!g_demoEnv.bEnumerateOnly) {
                DemoLogf("Opening Gimle board#%d\n", nBoardNo);
                DemoOpenGimleBoard(nBoardNo);
              }
              g_coGimleBrds++;
            }
            break;
            
          case OTX_DEVICE_THOR_PCI:
          case OTX_DEVICE_THOR_2_PCI_PLUS:
          case OTX_DEVICE_THOR_2_PCI_EXPRESS:
          case OTX_DEVICE_THOR_8_PCI:
          case OTX_DEVICE_THOR_8_PCI_PLUS:
          case OTX_DEVICE_THOR_4_PCI_PLUS:
            {
              OTX_UINT32 nSubSystemId;
              OTX_UINT32 nBoardNo;
              char szSerNo[128];
              
              // Read serial number from the board
              DemoGetBoardInformation( hParentDevice, hChildDevice, eDevType, &nSubSystemId, &nBoardNo, szSerNo);
#if 1              
              if (!strcmp(szSerNo, "T2PPX121201049")) {
                DemoLogf("********** Just temporary... Skipping Thor board serno %s\n", szSerNo);
                break;
              }
#endif
              if (!g_demoEnv.bEnumerateOnly) {
                DemoLogf("Opening Thor board#%d\n", nBoardNo);
                DemoOpenThorBoard(eDevType, nBoardNo);
              }
              g_coThorBrds++;
            }
            break;
          default:
            // Do nothing
            break;
          }
        } else {
          // Do nothing
          // DemoLogf("\nSkipped eDevType = %d (0x%X)\n\n", eDevType, eDevType);
        }
      } else {
        // This should not happen
        DemoLogf("Internal Error: queried a handle value but it still didn't work??? Should not happen!\n");
      }
    }
  }

  if (g_demoEnv.bEnumerateOnly) {
    exit(0);
  }

}



/*******************************************************************************

  @func

  Open one Physical Gimle Board Device
  
*******************************************************************************/
static void DemoOpenGimleBoard(OTX_INT32 nGimleBoardNo)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;
  OTX_INT32  iBrd = g_coGimleBrds;
  
  //********* Gimle *********
  result = OtxDrvOpenPhysicalDevice(0, OTX_DEVICE_GIMLE_16_PCI_PLUS, nGimleBoardNo, nGimleBoardNo, g_hEventQueue, &g_hPhysDevsGimle[iBrd].hBoard); 
  if (!OTX_SUCCESS(result)) {
    result = OtxDrvOpenPhysicalDevice(0, OTX_DEVICE_GIMLE_16_PCIE, nGimleBoardNo, nGimleBoardNo, g_hEventQueue, &g_hPhysDevsGimle[iBrd].hBoard); DemoCheckRc(result);
  }
  
  
  // On the Gimle-16-PCI-Plus board the E1T1 devices are hosted by four quad E1T1 devices. Open those 4 devices.
  for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
    DemoLogf("Open Gimle OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevsGimle[iBrd].hBoard, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevsGimle[iBrd].hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
  }
  
  // Get handles to T1/E1 line interface devices on the Gimle board
  for (i = 0; i < DEMO_CO_GIMLE_LI; i++) {
    nQuadE1T1 = i/4;
    DemoLogf("Open Gimle OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevsGimle[iBrd].hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i + iBrd*DEMO_CO_GIMLE_LI, g_hEventQueue, &g_hPhysDevsGimle[iBrd].hLi[i]); DemoCheckRc(result);
  }
  
  // Initialize E1 Configuration Data
  g_GimleT1E1CfgData.e1CfgData.eTxLineCode                = g_demoEnv.eGimleE1LineCode;
  g_GimleT1E1CfgData.e1CfgData.eRxLineCode                = g_demoEnv.eGimleE1LineCode;
  g_GimleT1E1CfgData.e1CfgData.eTxFrameFormat             = g_demoEnv.eGimleE1FrameFormat;
  g_GimleT1E1CfgData.e1CfgData.eRxFrameFormat             = g_demoEnv.eGimleE1FrameFormat;
  g_GimleT1E1CfgData.e1CfgData.eAisDetection              = OTX_E1_AIS_ETS300233;
  g_GimleT1E1CfgData.e1CfgData.bfSiBits                   = 0x3;
  g_GimleT1E1CfgData.e1CfgData.bfSaBits                   = 0x1F;
  g_GimleT1E1CfgData.e1CfgData.bExtendedHDB3errorDetect   = OTX_FALSE;
  g_GimleT1E1CfgData.e1CfgData.bAutoRegainMultiframe      = OTX_FALSE;
  g_GimleT1E1CfgData.e1CfgData.bAutoResync                = OTX_TRUE;
  g_GimleT1E1CfgData.e1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
  g_GimleT1E1CfgData.e1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
  g_GimleT1E1CfgData.e1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  g_GimleT1E1CfgData.e1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  g_GimleT1E1CfgData.e1CfgData.nLineLength                = 10;
  g_GimleT1E1CfgData.e1CfgData.bHighTxPower               = OTX_FALSE;
  g_GimleT1E1CfgData.e1CfgData.bReceiveEqualizer          = g_demoEnv.bGimleReceiveEqualizer;
  
  // Initialize T1 Configuration Data
  g_GimleT1E1CfgData.t1CfgData.eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  g_GimleT1E1CfgData.t1CfgData.eTxLineCode                = g_demoEnv.eGimleT1LineCode;
  g_GimleT1E1CfgData.t1CfgData.eRxLineCode                = g_demoEnv.eGimleT1LineCode;
  g_GimleT1E1CfgData.t1CfgData.eFrameFormat               = g_demoEnv.eGimleT1FrameFormat;
  g_GimleT1E1CfgData.t1CfgData.bCRC6Enable                = OTX_TRUE;
  g_GimleT1E1CfgData.t1CfgData.eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  g_GimleT1E1CfgData.t1CfgData.eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  g_GimleT1E1CfgData.t1CfgData.bAutoResync                = OTX_TRUE;
  g_GimleT1E1CfgData.t1CfgData.bAutoCrcBitInversion       = OTX_TRUE;
  g_GimleT1E1CfgData.t1CfgData.eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  g_GimleT1E1CfgData.t1CfgData.bAutoRemoteAlarm           = OTX_TRUE;
  g_GimleT1E1CfgData.t1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  g_GimleT1E1CfgData.t1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  g_GimleT1E1CfgData.t1CfgData.nLineLength                = 10;
  g_GimleT1E1CfgData.t1CfgData.bHighTxPower               = OTX_FALSE;
  g_GimleT1E1CfgData.t1CfgData.bReceiveEqualizer          = g_demoEnv.bGimleReceiveEqualizer;
  g_GimleT1E1CfgData.t1CfgData.bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
  
  // Configure Transceivers
  for (i = 0; i < DEMO_CO_GIMLE_LI; i++) {
    
    if (g_demoEnv.eGimleT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevsGimle[iBrd].hLi[i], &(g_GimleT1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevsGimle[iBrd].hLi[i], &(g_GimleT1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevsGimle[iBrd].hLi[i], g_demoEnv.eGimleT1E1Mode); DemoCheckRc(result);
    
    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevsGimle[iBrd].hLi[i], g_demoEnv.eGimleTermMode); DemoCheckRc(result);
    
    // Set receive threshold
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevsGimle[iBrd].hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nGimleRxThreshold); DemoCheckRc(result);
    
    // Set Monitor (-20dB) mode
    result = OtxT1E1SetMonitorMode(g_hPhysDevsGimle[iBrd].hLi[i], g_demoEnv.bGimleMonitorMode); DemoCheckRc(result);
    
    // Turning front end Amplifier ON/OFF      
    result = OtxDrvSetAttributeValueBool(g_hPhysDevsGimle[iBrd].hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bGimleAmplifier); DemoCheckRc(result);
    
    // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
    result = OtxT1E1XmitPinControl(g_hPhysDevsGimle[iBrd].hLi[i], OTX_FALSE); DemoCheckRc(result);
    
  }
  
  // Enable the Gimle board
  result = OtxDrvEnable(g_hPhysDevsGimle[iBrd].hBoard,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  
  // Enable Quad E1T1 devices
  for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
    DemoLogf("Enable Gimle OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
    result = OtxDrvEnable(g_hPhysDevsGimle[iBrd].hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
  
  // Enable Lis
  for (i = 0; i < DEMO_CO_GIMLE_LI; i++) {
    result = OtxDrvEnable(g_hPhysDevsGimle[iBrd].hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  DemoLogf("Connecting board to CTBUS#%d...\n", 0);
  result = OtxDrvConnectBoard(g_hCtBus, g_hPhysDevsGimle[iBrd].hBoard); DemoCheckRc(result);


}
 
  
/*******************************************************************************

  @func
    Open one Physical Thor Board Devices

*******************************************************************************/
static void DemoOpenThorBoard(OTX_INT32 eThorBoardType, OTX_INT32 nThorBoardNo)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;
  OTX_INT32  iBrd = g_coThorBrds;

  // Connect to the Thor board
  switch (eThorBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    g_hPhysDevsThor[iBrd].coThorLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    g_hPhysDevsThor[iBrd].coThorLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    g_hPhysDevsThor[iBrd].coThorLi = 2;
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    g_hPhysDevsThor[iBrd].coThorLi = 8;
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    g_hPhysDevsThor[iBrd].coThorLi = 8;
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    g_hPhysDevsThor[iBrd].coThorLi = 4;
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", eThorBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) eThorBoardType, nThorBoardNo, nThorBoardNo+100, g_hEventQueue, &g_hPhysDevsThor[iBrd].hBoard); DemoCheckRc(result);
  
  switch (eThorBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On the Thor-2-PCI types of boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_hPhysDevsThor[iBrd].coThorLi; i++) {
      DemoLogf("Open Thor OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor[iBrd].hBoard, OTX_DEVICE_LI_T1E1, i, (iBrd+1)*100+i, g_hEventQueue, &g_hPhysDevsThor[iBrd].hLi[i]); DemoCheckRc(result);
    }
    break;
    
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // On the Thor-8-PCI board the E1T1 devices are hosted by two quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open Thor OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor[iBrd].hBoard, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1+100, g_hEventQueue, &g_hPhysDevsThor[iBrd].hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }
    
    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_hPhysDevsThor[iBrd].coThorLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open Thor OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor[iBrd].hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, (iBrd+1)*100+i, g_hEventQueue, &g_hPhysDevsThor[iBrd].hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    // On the Thor-4-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Open Thor OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor[iBrd].hBoard, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1+100, g_hEventQueue, &g_hPhysDevsThor[iBrd].hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }
    
    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_hPhysDevsThor[iBrd].coThorLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open Thor OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor[iBrd].hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, (iBrd+1)*100+i, g_hEventQueue, &g_hPhysDevsThor[iBrd].hLi[i]); DemoCheckRc(result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", eThorBoardType);
    exit(1);
    break;
  }
  
  
  // Initialize E1 Configuration Data
  g_ThorT1E1CfgData.e1CfgData.eTxLineCode                = g_demoEnv.eThorE1LineCode;
  g_ThorT1E1CfgData.e1CfgData.eRxLineCode                = g_demoEnv.eThorE1LineCode;
  g_ThorT1E1CfgData.e1CfgData.eTxFrameFormat             = g_demoEnv.eThorE1FrameFormat;
  g_ThorT1E1CfgData.e1CfgData.eRxFrameFormat             = g_demoEnv.eThorE1FrameFormat;
  g_ThorT1E1CfgData.e1CfgData.eAisDetection              = OTX_E1_AIS_ETS300233;
  g_ThorT1E1CfgData.e1CfgData.bfSiBits                   = 0x3;
  g_ThorT1E1CfgData.e1CfgData.bfSaBits                   = 0x1F;
  g_ThorT1E1CfgData.e1CfgData.bExtendedHDB3errorDetect   = OTX_FALSE;
  g_ThorT1E1CfgData.e1CfgData.bAutoRegainMultiframe      = OTX_FALSE;
  g_ThorT1E1CfgData.e1CfgData.bAutoResync                = OTX_TRUE;
  g_ThorT1E1CfgData.e1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
  g_ThorT1E1CfgData.e1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
  g_ThorT1E1CfgData.e1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  g_ThorT1E1CfgData.e1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  g_ThorT1E1CfgData.e1CfgData.nLineLength                = 10;
  g_ThorT1E1CfgData.e1CfgData.bHighTxPower               = OTX_FALSE;
  g_ThorT1E1CfgData.e1CfgData.bReceiveEqualizer          = g_demoEnv.bThorReceiveEqualizer;
  
  // Initialize T1 Configuration Data
  g_ThorT1E1CfgData.t1CfgData.eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  g_ThorT1E1CfgData.t1CfgData.eTxLineCode                = g_demoEnv.eThorT1LineCode;
  g_ThorT1E1CfgData.t1CfgData.eRxLineCode                = g_demoEnv.eThorT1LineCode;
  g_ThorT1E1CfgData.t1CfgData.eFrameFormat               = g_demoEnv.eThorT1FrameFormat;
  g_ThorT1E1CfgData.t1CfgData.bCRC6Enable                = OTX_TRUE;
  g_ThorT1E1CfgData.t1CfgData.eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  g_ThorT1E1CfgData.t1CfgData.eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  g_ThorT1E1CfgData.t1CfgData.bAutoResync                = OTX_TRUE;
  g_ThorT1E1CfgData.t1CfgData.bAutoCrcBitInversion       = OTX_TRUE;
  g_ThorT1E1CfgData.t1CfgData.eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  g_ThorT1E1CfgData.t1CfgData.bAutoRemoteAlarm           = OTX_TRUE;
  g_ThorT1E1CfgData.t1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  g_ThorT1E1CfgData.t1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  g_ThorT1E1CfgData.t1CfgData.nLineLength                = 10;
  g_ThorT1E1CfgData.t1CfgData.bHighTxPower               = OTX_FALSE;
  g_ThorT1E1CfgData.t1CfgData.bReceiveEqualizer          = g_demoEnv.bThorReceiveEqualizer;
  g_ThorT1E1CfgData.t1CfgData.bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
  
  
  
  // Configure Transceivers
  for (i = 0; i < g_hPhysDevsThor[iBrd].coThorLi; i++) {
    
    if (g_demoEnv.eThorT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevsThor[iBrd].hLi[i], &(g_ThorT1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevsThor[iBrd].hLi[i], &(g_ThorT1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevsThor[iBrd].hLi[i], g_demoEnv.eThorT1E1Mode); DemoCheckRc(result);
    
    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevsThor[iBrd].hLi[i], g_demoEnv.eThorTermMode); DemoCheckRc(result);
    
    // Set receive threshold
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevsThor[iBrd].hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nThorRxThreshold); DemoCheckRc(result);
    
    // Set Monitor (-20dB) mode
    result = OtxT1E1SetMonitorMode(g_hPhysDevsThor[iBrd].hLi[i], g_demoEnv.bThorMonitorMode); DemoCheckRc(result);
    
    // Turning front end Amplifier ON/OFF      
    result = OtxDrvSetAttributeValueBool(g_hPhysDevsThor[iBrd].hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bThorAmplifier); DemoCheckRc(result);
    
    // We will use the transmit pins of the Lis in this application. Don't Tri-state the transmit pins
    result = OtxT1E1XmitPinControl(g_hPhysDevsThor[iBrd].hLi[i], OTX_TRUE); DemoCheckRc(result);
    
  }
  
  // Enable Board
  result = OtxDrvEnable(g_hPhysDevsThor[iBrd].hBoard,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  
  // Enable Quad E1T1 device if board has this device
  switch (eThorBoardType) {
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    for (nQuadE1T1 = 0; nQuadE1T1 < ((g_hPhysDevsThor[iBrd].coThorLi+3)/4); nQuadE1T1++) {
      DemoLogf("Enable Thor OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevsThor[iBrd].hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;
  default:
    // Nothing
    break;
  }
  
  for (i = 0; i < g_hPhysDevsThor[iBrd].coThorLi; i++) {
    result = OtxDrvEnable(g_hPhysDevsThor[iBrd].hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
  
  DemoLogf("Connecting board to CTBUS#%d...\n", 0);
  result = OtxDrvConnectBoard(g_hCtBus, g_hPhysDevsThor[iBrd].hBoard); DemoCheckRc(result);


}


/*******************************************************************************

  @func
    Set the clock source for all boards

*******************************************************************************/
static void DemoSetClockSource()
{
  OTX_RESULT result;
  OTX_INT32  iBrd;
  char szClockSource[64];

  DemoLogf("Gimle board#%d is the clock source for the system\n", g_demoEnv.nGimleClkSourceBrd);

  iBrd = g_demoEnv.nGimleClkSourceBrd;

  switch (g_demoEnv.eGimleClkSource) {
  case OTX_CLOCK_SOURCE_INTERNAL:
    strcpy(szClockSource, "Internal");
    break;
  case OTX_CLOCK_SOURCE_LOCAL_0:
  case OTX_CLOCK_SOURCE_LOCAL_1:
  case OTX_CLOCK_SOURCE_LOCAL_2:
  case OTX_CLOCK_SOURCE_LOCAL_3:
  case OTX_CLOCK_SOURCE_LOCAL_4:
  case OTX_CLOCK_SOURCE_LOCAL_5:
  case OTX_CLOCK_SOURCE_LOCAL_6:
  case OTX_CLOCK_SOURCE_LOCAL_7:
  case OTX_CLOCK_SOURCE_LOCAL_8:
  case OTX_CLOCK_SOURCE_LOCAL_9:
  case OTX_CLOCK_SOURCE_LOCAL_10:
  case OTX_CLOCK_SOURCE_LOCAL_11:
  case OTX_CLOCK_SOURCE_LOCAL_12:
  case OTX_CLOCK_SOURCE_LOCAL_13:
  case OTX_CLOCK_SOURCE_LOCAL_14:
  case OTX_CLOCK_SOURCE_LOCAL_15:
    sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eGimleClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
    break;
  default:
    strcpy(szClockSource, "Unknown");
    break;
  }
  
  DemoLogf("Clock Source Gimle Board#%d: %s\n", iBrd, szClockSource);
  
  // Set clock source for this board (source for the H.100 A-clocks)
  result = OtxBrdSetClocks(g_hPhysDevsGimle[iBrd].hBoard, OTX_CLOCK_MODE_H100_ETCF, g_demoEnv.eGimleClkSource); DemoCheckRc(result);


  // All other Gimle boards will slave off the H.100 A-clocks
  for (iBrd = 0; iBrd < g_coGimleBrds; iBrd++) {
    if (iBrd == g_demoEnv.nGimleClkSourceBrd) continue; // We already set the clock source for this board above

    // Set clock source for this board
    DemoLogf("Clock Source Gimle Board#%d: %s\n", iBrd, "H.100 A-clocks");
    result = OtxBrdSetClocks(g_hPhysDevsGimle[iBrd].hBoard, OTX_CLOCK_MODE_H100_ETCF, OTX_CLOCK_SOURCE_A_CLOCKS_ETCF); DemoCheckRc(result);
  }

  // All other Thor boards will also slave off the H.100 A-clocks
  for (iBrd = 0; iBrd < g_coThorBrds; iBrd++) {
    // Set clock source for this board
    DemoLogf("Clock Source Thor Board#%d: %s\n", iBrd, "H.100 A-clocks");
    result = OtxBrdSetClocks(g_hPhysDevsThor[iBrd].hBoard, OTX_CLOCK_MODE_H100_ETCF, OTX_CLOCK_SOURCE_A_CLOCKS_ETCF); DemoCheckRc(result);
  }
  

}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

#define DEMO_EVENT_DATA_BUF_SIZE 10

/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    DemoLogf("Event signalled\n");

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) 
{
  short  i;
  
  // Set reasonable defaults before reading from command line
  envP->pszInputStr            = NULL;
  envP->bExitOnError           = OTX_FALSE;
  envP->bEnumerateOnly         = OTX_FALSE;
  
  envP->nGimleClkSourceBrd     = 0;
  envP->eGimleClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->eGimleT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bGimleReceiveEqualizer = OTX_TRUE;
  envP->eGimleTermMode         = OTX_T1E1_TERM_300_OHM;
  envP->bGimleMonitorMode      = OTX_TRUE;
  envP->bGimleAmplifier        = OTX_TRUE;
  envP->nGimleRxThreshold      = 10; // in mV
  envP->eGimleT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eGimleT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eGimleE1LineCode       = OTX_E1_LC_HDB3;
  envP->eGimleE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  
  envP->eThorT1E1Mode          = OTX_T1E1_MODE_E1;
  envP->bThorReceiveEqualizer  = OTX_FALSE;
  envP->eThorTermMode          = OTX_T1E1_TERM_100_OHM;
  envP->bThorMonitorMode       = OTX_FALSE;
  envP->bThorAmplifier         = OTX_FALSE;
  envP->nThorRxThreshold       = 500; // in mV
  envP->eThorT1LineCode        = OTX_T1_LC_B8ZS;
  envP->eThorT1FrameFormat     = OTX_T1_FF_ESF;
  envP->eThorE1LineCode        = OTX_E1_LC_HDB3;
  envP->eThorE1FrameFormat     = OTX_E1_FF_CRC4_MULTIFRAME;
  
  
  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  strcpy(envP->conFileName, envP->cwd);
  strcat(envP->conFileName, DEMO_DEFAULT_CON_FILE);

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file name
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'C') {   // Connection file name
          strcpy(envP->conFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'E') {   // Enumerate Only
          envP->bEnumerateOnly = OTX_TRUE;
          
          // ****** Gimle Parameters ******
        } else if (*(argv[i]+1) == 'G') {   // Gimle Parameters
          if (*(argv[i]+2) == 'K') {   // Clock Source
            if ( (*(argv[i]+3) == '1') && (*(argv[i]+4) == '0') ) {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
            } else if ( (*(argv[i]+3) == '1') && (*(argv[i]+4) == '1') ) {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
            } else if ( (*(argv[i]+3) == '1') && (*(argv[i]+4) == '2') ) {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
            } else if ( (*(argv[i]+3) == '1') && (*(argv[i]+4) == '3') ) {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
            } else if ( (*(argv[i]+3) == '1') && (*(argv[i]+4) == '4') ) {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
            } else if ( (*(argv[i]+3) == '1') && (*(argv[i]+4) == '5') ) {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
            } else if (*(argv[i]+3) == 'I') {   // Internal
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            } else if (*(argv[i]+3) == '0') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
            } else if (*(argv[i]+3) == '1') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
            } else if (*(argv[i]+3) == '2') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
            } else if (*(argv[i]+3) == '3') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
            } else if (*(argv[i]+3) == '4') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
            } else if (*(argv[i]+3) == '5') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
            } else if (*(argv[i]+3) == '6') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
            } else if (*(argv[i]+3) == '7') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
            } else if (*(argv[i]+3) == '8') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
            } else if (*(argv[i]+3) == '9') {   
              envP->eGimleClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
            } else {
              DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
              DemoHelpPanel();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'N') {
            // RxThreshold (in mV)
            envP->nGimleClkSourceBrd = strtoul(argv[i] + 3, NULL, 0);
          } else if (*(argv[i]+2) == 'M') {   // Li Mode
            if (*(argv[i]+3) == 'E') {   // E1 
              envP->eGimleT1E1Mode    = OTX_T1E1_MODE_E1;
            } else if (*(argv[i]+3) == 'T') {   // T1
              envP->eGimleT1E1Mode    = OTX_T1E1_MODE_T1;
            } else {
              DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'E') {
            // Longhaul mode (ReceiveEqualizer)
            if (*(argv[i]+3) == 'Y') {   // Yes 
              envP->bGimleReceiveEqualizer = OTX_TRUE;
            } else if (*(argv[i]+3) == 'N') {   // NO
              envP->bGimleReceiveEqualizer = OTX_FALSE;
            } else {
              DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'Y') {
            // TermMode
            if ( *(argv[i]+3) == '1') {
              envP->eGimleTermMode = OTX_T1E1_TERM_100_OHM;
            } else if ( *(argv[i]+3) == '7') {
              envP->eGimleTermMode = OTX_T1E1_TERM_75_OHM;
            } else if ( *(argv[i]+3) == 'Z') {
              envP->eGimleTermMode = OTX_T1E1_TERM_HIGHZ;
            } else if ( *(argv[i]+3) == 'S') {
              envP->eGimleTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
            } else if ( *(argv[i]+3) == '3') {
              envP->eGimleTermMode = OTX_T1E1_TERM_300_OHM;
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == '2') {
            // Use MonitorMode  = ON
            if (*(argv[i]+3) == 'Y') {   // Yes 
              envP->bGimleMonitorMode = OTX_TRUE;
            } else if (*(argv[i]+3) == 'N') {   // NO
              envP->bGimleMonitorMode = OTX_FALSE;
            } else {
              DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'A') {
            // Use Front End Amplifier = ON
            if (*(argv[i]+3) == 'Y') {   // Yes 
              envP->bGimleAmplifier = OTX_TRUE;
            } else if (*(argv[i]+3) == 'N') {   // NO
              envP->bGimleAmplifier = OTX_FALSE;
            } else {
              DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'H') {
            // RxThreshold (in mV)
            envP->nGimleRxThreshold = strtoul(argv[i] + 3, NULL, 0);
          } else if (*(argv[i]+2) == 'J') {   // T1 Line Code
            if ( *(argv[i]+3) == 'A') {
              envP->eGimleT1LineCode = OTX_T1_LC_AMI; // AMI
            } else if ( *(argv[i]+3) == 'B') {
              envP->eGimleT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'F') {   // T1 Frame format
            if ( *(argv[i]+3) == 'E') {
              envP->eGimleT1FrameFormat = OTX_T1_FF_ESF; // ESF
            } else if ( *(argv[i]+3) == '4') {
              envP->eGimleT1FrameFormat = OTX_T1_FF_F4; 
            } else if ( *(argv[i]+3) == 'S') {
              envP->eGimleT1FrameFormat = OTX_T1_FF_F12; // SF
            } else if ( *(argv[i]+3) == '9') {
              envP->eGimleT1FrameFormat = OTX_T1_FF_F72; // SLC96
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'U') {   // E1 Line Code
            if ( *(argv[i]+3) == 'A') {
              envP->eGimleE1LineCode = OTX_E1_LC_AMI; // AMI
            } else if ( *(argv[i]+3) == 'H') {
              envP->eGimleE1LineCode = OTX_E1_LC_HDB3; // HDB3
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'G') {   // E1 Frame format
            if ( *(argv[i]+3) == 'D') {
              envP->eGimleE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
            } else if ( *(argv[i]+3) == 'C') {
              envP->eGimleE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
            } else if ( *(argv[i]+3) == 'G') {
              envP->eGimleE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else {
            DemoLogf("Error: Unknown Gimle command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
          // ****** Thor Parameters ******
        } else if (*(argv[i]+1) == 'T') {   // Thor Parameters
          if (*(argv[i]+2) == 'M') {   // Li Mode
            if (*(argv[i]+3) == 'E') {   // E1 
              envP->eThorT1E1Mode    = OTX_T1E1_MODE_E1;
            } else if (*(argv[i]+3) == 'T') {   // T1
              envP->eThorT1E1Mode    = OTX_T1E1_MODE_T1;
            } else {
              DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'E') {
            // Longhaul mode (ReceiveEqualizer)
            if (*(argv[i]+3) == 'Y') {   // Yes 
              envP->bThorReceiveEqualizer = OTX_TRUE;
            } else if (*(argv[i]+3) == 'N') {   // NO
              envP->bThorReceiveEqualizer = OTX_FALSE;
            } else {
              DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'Y') {
            // TermMode
            if ( *(argv[i]+3) == '1') {
              envP->eThorTermMode = OTX_T1E1_TERM_100_OHM;
            } else if ( *(argv[i]+3) == '7') {
              envP->eThorTermMode = OTX_T1E1_TERM_75_OHM;
            } else if ( *(argv[i]+3) == 'Z') {
              envP->eThorTermMode = OTX_T1E1_TERM_HIGHZ;
            } else if ( *(argv[i]+3) == 'S') {
              envP->eThorTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
            } else if ( *(argv[i]+3) == '3') {
              envP->eThorTermMode = OTX_T1E1_TERM_300_OHM;
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == '2') {
            // Use MonitorMode  = ON
            if (*(argv[i]+3) == 'Y') {   // Yes 
              envP->bThorMonitorMode = OTX_TRUE;
            } else if (*(argv[i]+3) == 'N') {   // NO
              envP->bThorMonitorMode = OTX_FALSE;
            } else {
              DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'A') {
            // Use Front End Amplifier = ON
            if (*(argv[i]+3) == 'Y') {   // Yes 
              envP->bThorAmplifier = OTX_TRUE;
            } else if (*(argv[i]+3) == 'N') {   // NO
              envP->bThorAmplifier = OTX_FALSE;
            } else {
              DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'H') {
            // RxThreshold (in mV)
            envP->nThorRxThreshold = strtoul(argv[i]+3, NULL, 0);
          } else if (*(argv[i]+2) == 'J') {   // T1 Line Code
            if ( *(argv[i]+3) == 'A') {
              envP->eThorT1LineCode = OTX_T1_LC_AMI; // AMI
            } else if ( *(argv[i]+3) == 'B') {
              envP->eThorT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'F') {   // T1 Frame format
            if ( *(argv[i]+3) == 'E') {
              envP->eThorT1FrameFormat = OTX_T1_FF_ESF; // ESF
            } else if ( *(argv[i]+3) == '4') {
              envP->eThorT1FrameFormat = OTX_T1_FF_F4; 
            } else if ( *(argv[i]+3) == 'S') {
              envP->eThorT1FrameFormat = OTX_T1_FF_F12; // SF
            } else if ( *(argv[i]+3) == '9') {
              envP->eThorT1FrameFormat = OTX_T1_FF_F72; // SLC96
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'U') {   // E1 Line Code
            if ( *(argv[i]+3) == 'A') {
              envP->eThorE1LineCode = OTX_E1_LC_AMI; // AMI
            } else if ( *(argv[i]+3) == 'H') {
              envP->eThorE1LineCode = OTX_E1_LC_HDB3; // HDB3
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else if (*(argv[i]+2) == 'G') {   // E1 Frame format
            if ( *(argv[i]+3) == 'D') {
              envP->eThorE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
            } else if ( *(argv[i]+3) == 'C') {
              envP->eThorE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
            } else if ( *(argv[i]+3) == 'G') {
              envP->eThorE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
            } else {
              DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
              DemoHelpPanel();
              OsaConsoleExit();
              exit(1);
            }
          } else {
            DemoLogf("Error: Unknown Thor command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
  
  
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf("Global switches:\n");
  DemoLogf(" /L<fileName>   : Log file name (default is %s)\n", DEMO_DEFAULT_LOG_FILE);
  DemoLogf(" /C<fileName>   : Timeslot connection file name (default is %s)\n", DEMO_DEFAULT_CON_FILE);
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /E             : Enumerate Gimle and Thor boards, then exit\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
  DemoLogf("Board specific switches:\n");
  DemoLogf(" /G<boardswitch> : Switch for Gimle board (see below)\n");
  DemoLogf(" /T<boardswitch> : Switch for Thor board (see below)\n");
  DemoLogf("    M{E|T}       : E1/T1 mode (E=E1 mode, T=T1 mode)\n");
  DemoLogf("    K{x}         : Clock Source (only for Gimle board), where\n");
  DemoLogf("                   x = 0 through 15 indicating which Li to recover the clock from or\n");
  DemoLogf("                   x = I Internal clock source (will always result in slips)\n");
  DemoLogf("    N<brdNo>     : Board number of clock source for system (Gimle only)\n");
  DemoLogf("    A{Y|N}       : Front End Amplifier (Y=ON, N=OFF)\n");
  DemoLogf("    E{Y|N}       : Receive Equalizer (Y=ON, N=OFF)\n");
  DemoLogf("    2{Y|N}       : -20dB MonitorMode (Y=ON, N=OFF)\n");
  DemoLogf("    H<threshold> : Receive Input Threshold in mV\n");
  DemoLogf("    Y{1|7|Z|S|3} : Termination Mode, where\n");
  DemoLogf("                    1 = 100/120 ohm balanced\n");
  DemoLogf("                    7 = 75 ohm unbalanced\n");
  DemoLogf("                    Z = High-impedance\n");
  DemoLogf("                    S = 75 ohm balanced\n");
  DemoLogf("                    3 = 300 ohm balanced (for attenuated signals)\n");
  DemoLogf("    F{N|E|S|9}   : T1 Frame Format, where\n");
  DemoLogf("                    N = Not framed/Unframed\n");
  DemoLogf("                    E = ESF (default)\n");
  DemoLogf("                    4 = F4\n");
  DemoLogf("                    S = SF [D3/D4] (F12)\n");
  DemoLogf("                    9 = SLC96 (F72)\n");
  DemoLogf("    J{A|B}       : T1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    B = B8ZS (default)\n");
  DemoLogf("    G{N|D|C|G}   : E1 Frame Format, where\n");
  DemoLogf("                    N = Not framed/Unframed\n");
  DemoLogf("                    D = Doubleframe\n");
  DemoLogf("                    C = CRC4 Multiframe (default)\n");
  DemoLogf("                    G = CRC4 Multiframe (G.706)\n");
  DemoLogf("    U{A|H}       : E1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    H = HDB3 (default)\n");
  DemoLogf("\nExample:\n");
  DemoLogf("/TMT /GK15 /GN0\n");
  DemoLogf("/TMT   We want to use T1 mode for the Thor board\n");
  DemoLogf("/GK15  We want to recover the clock from Gimle's Li15\n");
  DemoLogf("/GN0   We want to recover the clock from Gimle board number 0\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);

    DemoExitQuestion();

  }
  
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
   Returns the string representing the device type for a specific device, automatically chooses between logical and physical devices
*******************************************************************************/
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE hDevice)
{
  static char szBuf[256];
  OTX_RESULT otxRc;
  
  otxRc = OtxDrvGetAttributeValueSz(hDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
  if (!OTX_SUCCESS(otxRc)) {
      DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hDevice);
      return "*** Error ***";
  } else {
    return szBuf;
  }
}


/*******************************************************************************
  @func
   Returns the device no for a specific device
*******************************************************************************/
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_UINT32 nDevNo;
  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
    return nDevNo;
  } else {
    return -1;
  }
}

/*******************************************************************************
  @func
   Returns the device no of the host for a specific device
*******************************************************************************/
static signed int DemoGetNicDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_HANDLE hNicDevice;
  OTX_UINT32 nDevNo;

  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NIC, &hNicDevice))) {
    if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hNicDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
      return nDevNo;
    } else {
      return -1;
    }
  } else {
    return -1;
  }

}




/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintPipeConnections(
  OTX_HANDLE hPipe
)
{
  OTX_RESULT result;
  OTX_UINT32 nCapacity;
  OTX_UINT32 nInHandle;
  OTX_UINT32 nOutHandle;
  OTX_UINT32 nInHighway;
  OTX_UINT32 nOutHighway;
  OTX_UINT32 cbReturned;
  OTX_UINT8 sqInTimeSlotMask[128];
  OTX_UINT8 sqOutTimeSlotMask[128];
  OTX_UINT32 n;
    
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_CAPACITY, &nCapacity); DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_IN_CONNECTED, &nInHandle); DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_OUT_CONNECTED, &nOutHandle); DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_IN_HIGHWAY, &nInHighway); DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_OUT_HIGHWAY, &nOutHighway); DemoCheckRc(result);
  result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_IN_CONN_MASK, sqInTimeSlotMask, 128, &cbReturned); DemoCheckRc(result);
  result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_OUT_CONN_MASK, sqOutTimeSlotMask, 128, &cbReturned); DemoCheckRc(result);

  DemoLogf("Connection info for pipe 0x%X (Capacity:%d)\n", hPipe, nCapacity);

  DemoLogf("  Input: Brd:%d %s[%d] Hwy:%d ", DemoGetNicDeviceNo((OTX_HANDLE)nInHandle), DemoGetDeviceTypeString((OTX_HANDLE)nInHandle), DemoGetDeviceNo((OTX_HANDLE)nInHandle), nInHighway);

  DemoLogf("TS: ");
  for (n = 0; n < 32; n++) {
    if (sqInTimeSlotMask[n]) {
      DemoLogf("%d ", n);
    }
  }
  DemoLogf("\n");

  DemoLogf("  Output: Brd:%d %s[%d] Hwy:%d ", DemoGetNicDeviceNo((OTX_HANDLE)nOutHandle), DemoGetDeviceTypeString((OTX_HANDLE)nOutHandle), DemoGetDeviceNo((OTX_HANDLE)nOutHandle), nOutHighway);

  DemoLogf("TS: ");
  for (n = 0; n < 32; n++) {
    if (sqOutTimeSlotMask[n]) {
      DemoLogf("%d ", n);
    }
  }
  DemoLogf("\n");

  // CT Bus connection info
  {
    OtxPipeCtBusConnectionS *psqCtBusConnectPoints;
    OTX_UINT32  cbCtBusConnectPoints;  
    OTX_UINT32  coCtBusConnectPoints;  
    OTX_UINT32  cbAttribValue;   // Bytes actually used in psqCtBusConnectPoints

    result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_CT_BUS_CONNECTIONS_COUNT, &coCtBusConnectPoints); DemoCheckRc(result);
    cbCtBusConnectPoints = (coCtBusConnectPoints * sizeof(OtxPipeCtBusConnectionS));

    psqCtBusConnectPoints = (OtxPipeCtBusConnectionS *) malloc(cbCtBusConnectPoints * sizeof(char));
    assert(psqCtBusConnectPoints);

    result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_CT_BUS_CONNECTIONS_DETAILS, psqCtBusConnectPoints, cbCtBusConnectPoints, &cbAttribValue); DemoCheckRc(result);
    assert(cbAttribValue == cbCtBusConnectPoints);

    if (coCtBusConnectPoints) {
      for (n = 0; n < coCtBusConnectPoints; n++) {
        DemoLogf("  Sourcing CT Bus:%d, CT Highway:%d, CT TS:%d\n", psqCtBusConnectPoints[n].nCtBus, psqCtBusConnectPoints[n].nHighway, psqCtBusConnectPoints[n].nTimeslot );
      }
    } else {
      DemoLogf("  No sourced CT Bus connections\n");
    }

    free(psqCtBusConnectPoints);
  }

    
}



/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentLi = 0;
  static OTX_INT32      iCurrentGimleBrd = 0;
  OTX_RESULT result;

  DemoLogf("\nKey %c pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" n   Increment current Gimle Li (currently Li#%d selected)\n", nCurrentLi);
    DemoLogf(" N   Decrement current Gimle Li (currently Li#%d selected)\n", nCurrentLi);
    DemoLogf(" >   Increment current Gimle Brd (currently Brd#%d selected)\n", iCurrentGimleBrd);
    DemoLogf(" <   Decrement current Gimle Brd (currently Brd#%d selected)\n", iCurrentGimleBrd);
    DemoLogf(" l   Query Line Status of all Lis (Gimle and Thor)\n");
    DemoLogf(" L   Query Extended Line Status of all Lis (Gimle and Thor)\n");
    DemoLogf(" F   Query T1/E1 span frequency of the current clock source (Gimle)\n");
    DemoLogf(" z   Set T1/E1 Line Termination for current Gimle Li\n");
    DemoLogf(" &   Read receive equalizer status (T1 mode only) (Gimle)\n");
    DemoLogf(" J   Set Receive Input Threshold for current Gimle Li\n");
    DemoLogf(" m/M Enable/Disable Monitor mode for current Gimle Li (for -20dB attenuation)\n");
    DemoLogf(" g/G Turn front end Amplifier On/Off for current Gimle Li (for -30dB attenuation)\n");
    DemoLogf(" f   Force Resynchronization on all Gimle Lis\n");
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" [   Measure T1/E1 span voltage for all Lis (Gimle)\n");
    break;

    // E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 nMilliVolts;
      OTX_INT32 nLi;
            
      DemoLogf("Measure T1/E1 Span Voltage for Gimle Lis: \n");
      for (nLi = 0; nLi < DEMO_CO_GIMLE_LI; nLi++) {
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nLi], OTX_ATTR_T1E1_LINE_VOLTAGE, &nMilliVolts); DemoCheckRc(result);
        DemoLogf("  Li%d:  Voltage = %2.3f V   (%2.1f dB)\n", nLi, (double)nMilliVolts/1000, (nMilliVolts) ? log10((double)nMilliVolts/3000)*20 : 0);
      }
      
    }
    break;

  case 'f':
    {
      DemoLogf("Forcing resynchronization on Gimle Li #%d... ", nCurrentLi);
      result = OtxT1E1ForceResynch(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi]); DemoCheckRc(result);
      DemoLogf("Done.");
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;
      OTX_INT32 iBrd;

      DemoLogf("Gimle (input spans):\n");
      for (iBrd = 0; iBrd < g_coGimleBrds; iBrd++) {
        for (nLi = 0; nLi < DEMO_CO_GIMLE_LI; nLi++) {
          DemoLogf("Gimle#%d: Line status of Li #%d is ", iBrd, nLi);
          result = OtxT1E1GetLayer1Status(g_hPhysDevsGimle[iBrd].hLi[nLi], &eLayer1Status); DemoCheckRc(result);
          DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
        }
      }
      DemoLogf("Thor (output spans):\n");
      for (iBrd = 0; iBrd < g_coThorBrds; iBrd++) {
        for (nLi = 0; nLi < g_hPhysDevsThor[iBrd].coThorLi; nLi++) {
          DemoLogf("Thor#%d: Line status of Li #%d is ", iBrd, nLi);
          result = OtxT1E1GetLayer1Status(g_hPhysDevsThor[iBrd].hLi[nLi], &eLayer1Status); DemoCheckRc(result);
          DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
        }
      }
    }
    break;

  case 'L':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;
      OTX_INT32 iBrd;

      DemoLogf("Gimle (input spans):\n");
      for (iBrd = 0; iBrd < g_coGimleBrds; iBrd++) {
        for (nLi = 0; nLi < DEMO_CO_GIMLE_LI; nLi++) {
          DemoLogf("Gimle#%d: Li #%d line is ", iBrd, nLi);
          result = OtxT1E1GetLayer1StatusEx(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nLi], &sLayer1Status); DemoCheckRc(result);
          
          DemoLogf("Alarms: ");
          if (sLayer1Status.bRedAlarm) {
            DemoLogf("R");
          } else {
            DemoLogf("-");
          }
          if (sLayer1Status.bBlueAlarm) {
            DemoLogf("B");
          } else {
            DemoLogf("-");
          }
          if (sLayer1Status.bYellowAlarm) {
            DemoLogf("Y");
          } else {
            DemoLogf("-");
          }
          
          if (sLayer1Status.bFrameError) {
            DemoLogf(" FrmErr: YES");
          } else {
            DemoLogf(" FrmErr: NO ");
          }
          
          if (sLayer1Status.bMultiFrameError) {
            DemoLogf(" MFrmErr: YES");
          } else {
            DemoLogf(" MFrmErr: NO ");
          }
          DemoLogf("\n");
        }
      }
      DemoLogf("Thor (output spans):\n");
      for (iBrd = 0; iBrd < g_coThorBrds; iBrd++) {
        for (nLi = 0; nLi < g_hPhysDevsThor[iBrd].coThorLi; nLi++) {
          DemoLogf("Thor#%d: Li #%d line is ", iBrd, nLi);
          result = OtxT1E1GetLayer1StatusEx(g_hPhysDevsThor[iBrd].hLi[nLi], &sLayer1Status); DemoCheckRc(result);
          
          DemoLogf("Alarms: ");
          if (sLayer1Status.bRedAlarm) {
            DemoLogf("R");
          } else {
            DemoLogf("-");
          }
          if (sLayer1Status.bBlueAlarm) {
            DemoLogf("B");
          } else {
            DemoLogf("-");
          }
          if (sLayer1Status.bYellowAlarm) {
            DemoLogf("Y");
          } else {
            DemoLogf("-");
          }
          
          if (sLayer1Status.bFrameError) {
            DemoLogf(" FrmErr: YES");
          } else {
            DemoLogf(" FrmErr: NO ");
          }
          
          if (sLayer1Status.bMultiFrameError) {
            DemoLogf(" MFrmErr: YES");
          } else {
            DemoLogf(" MFrmErr: NO ");
          }
          DemoLogf("\n");
        }
      }
    }
    break;


  case 'm':
    DemoLogf("Enabling Monitor (-20dB) mode for Gimle Brd#%d Li#%d \n", iCurrentGimleBrd, nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_TRUE); DemoCheckRc(result);
    DemoLogf("Done.\n");
    break;

  case 'M':
    DemoLogf("Disabling Monitor (-20dB) mode for Gimle Brd#%d Li#%d \n", iCurrentGimleBrd, nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_FALSE); DemoCheckRc(result);
    DemoLogf("Done.\n");
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;
      int c;

      DemoLogf("Set Line Termination mode for Gimle Brd#%d Li#%d:\n", iCurrentGimleBrd, nCurrentLi);
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      DemoLogf("4 = OTX_T1E1_TERM_300_OHM (for highly attenuated signals)\n");
      c = OsaConsoleGetCh();
      DemoLogf("'%c' pressed\n", c);
      switch (c - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      case 4:
        eTermMode = OTX_T1E1_TERM_300_OHM;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], eTermMode); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case 'g':
    {
      DemoLogf("Turning front end Amplifier ON for Gimle Brd#%d Li#%d \n", iCurrentGimleBrd, nCurrentLi);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, OTX_TRUE); DemoCheckRc(result);
      DemoLogf("Done.\n");
    }
    break;

  case 'G':
    {
      DemoLogf("Turning front end Amplifier OFF for Gimle Brd#%d Li#%d \n", iCurrentGimleBrd, nCurrentLi);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, OTX_FALSE); DemoCheckRc(result);
      DemoLogf("Done.\n");
    }
    break;


  case 'F':
    {
      OTX_UINT32 nFrequencyCount;

      DemoLogf("Querying frequency of current clock source for Gimle Brd#%d ...\n", iCurrentGimleBrd);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevsGimle[iCurrentGimleBrd].hBoard, OTX_ATTR_BRD_FREQUENCY_COUNTER, &nFrequencyCount); DemoCheckRc(result);

      if (OTX_SUCCESS(result)) {
        OTX_UINT32 nFrequency;

        if (g_demoEnv.eGimleT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nFrequency = (nFrequencyCount+1)*193/2048;
        } else {
          // E1 mode
          nFrequency = (nFrequencyCount+1)/8;
        }
        DemoLogf("Line frequency is %d Hz (nFrequencyCount=%d)\n", nFrequency, nFrequencyCount);
      }

      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case '&':
    {
      OTX_UINT32 readVal;
      OTX_INT32 nLi;
      char *pszEqStatus;
      
      for (nLi = 0; nLi < DEMO_CO_GIMLE_LI; nLi++) {
        result = OtxDrvReadReg8(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nLi], 115, &readVal); DemoCheckRc(result);
        switch ((readVal & 0xC0) >> 6) {
        case 0x00:
          pszEqStatus = "not valid, still adapting";
          break;
        case 0x01:
          pszEqStatus = "valid";
          break;
        case 0x02:
          pszEqStatus = "not valid";
          break;
        case 0x03:
          pszEqStatus = "valid but high noise floor";
          break;
        default:
          pszEqStatus = "unknown";
          break;
        }
        DemoLogf("Receive Equalizer of Gimle Brd#%d Li#%d: status=%s attenuation=%2.1fdB (0x%.2X)\n", iCurrentGimleBrd, nLi, pszEqStatus, (readVal & 0x1F)*1.4, readVal );
      }
      
    }
    break; 
    
  case 'J':
    {
      char strBuf[12];
      OTX_UINT32 nThresholdInMilliVolts;

      DemoLogf("Set Receive Input Threshold for Gimle Brd#%d Li%d \n", iCurrentGimleBrd, nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(result);

      DemoLogf("Threshold (%d mV): ", nThresholdInMilliVolts);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nThresholdInMilliVolts = strtoul(strBuf, NULL, 0);

        result = OtxDrvSetAttributeValueUint32(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, nThresholdInMilliVolts); DemoCheckRc(result);

        result = OtxDrvGetAttributeValueUint32(g_hPhysDevsGimle[iCurrentGimleBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(result);
        DemoLogf("Receive Input Threshold is changed to %d mV.\n", nThresholdInMilliVolts);
      } else {
        DemoLogf("The Receive Input Threshold remains unchanged.\n");
      }

    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'n':
    {
      // Increment nCurrentLi
      nCurrentLi++;

      if (nCurrentLi >= DEMO_CO_GIMLE_LI) {
        // Wrap 
        nCurrentLi = 0;
      }
      
      DemoLogf("Li#%d selected\n", nCurrentLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Decrement nCurrentLi
      nCurrentLi--;

      if (nCurrentLi < 0) {
        // Wrap
        nCurrentLi = DEMO_CO_GIMLE_LI - 1;
      }
      
      DemoLogf("Li#%d selected \n", nCurrentLi);
    }
    break;

  case '>':
    {
      // Increment iCurrentGimleBrd
      iCurrentGimleBrd++;

      if (iCurrentGimleBrd >= g_coGimleBrds) {
        // Wrap 
        iCurrentGimleBrd = 0;
      }
      
      DemoLogf("Gimle Brd#%d selected\n", iCurrentGimleBrd);
      
    }
    break;
    
    
  case '<':
    {
      // Decrement iCurrentGimleBrd
      iCurrentGimleBrd--;

      if (iCurrentGimleBrd < 0) {
        // Wrap
        iCurrentGimleBrd = g_coGimleBrds - 1;
      }
      
      DemoLogf("Gimle Brd#%d selected\n", iCurrentGimleBrd);
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


void DemoConnectDevicesHardcoded(
/*******************************************************************************

  @func
    
    Hardcode pipe connections for a specfic demo

    ODIN Test Setup (POC)-
    Two (-26dBm) input E1s)
    E1 No.1 (Li0-Rx pins) SL at TS 2(a) & 17(b) master clock source (TS#0)
    E1 No.2 (Li1-Rx pins) SL at TS 1(c) & 16(d) optional secondary clock source (TS#0)
    One (0dBm) output E1 (Li0-Tx pins)
    TS 1 is cross connect of E1#1(Li0)a
    TS 2 is cross connect of E1#1(Li0)b
    TS 3 is cross connect of E1#2(Li1)c
    TS 4 is cross connect of E1#2(Li1)d


*******************************************************************************/
  void
)
{
  OTX_INT32 nTs;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_HANDLE hPipe;
  OTX_INT32 iBrd = 0;

  DemoLogf(
    "ODIN Test Setup (POC)-\n"
    "Two (-26dBm) input E1s (LinkBit output)\n"
    "E1 No.1 (Li0-Rx pins) SL at TS 2(a) & 17(b) master clock source (TS#0)\n"
    "E1 No.2 (Li1-Rx pins) SL at TS 1(c) & 16(d) optional secondary clock source (TS#0)\n"
    "One (0dBm) output E1 (Li0-Tx pins) (GL input).\n"
    "TS 1 is cross connect of E1#1(Li0)a\n"
    "TS 2 is cross connect of E1#1(Li0)b\n"
    "TS 3 is cross connect of E1#2(Li1)c\n"
    "TS 4 is cross connect of E1#2(Li1)d\n"
    );
  
  
  // TS 1 is cross connect of E1#1(Li0)a
  result = OtxDrvCreatePipe(64, &hPipe);
  DemoCheckRc(result);

  nTs = 2;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hPhysDevsGimle[iBrd].hLi[0],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);
  
  nTs = 1;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor[iBrd].hLi[0],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);

  DemoLogf("Pipe for TS%d: \n", nTs); DemoPrintPipeConnections(hPipe);

  
  // TS 2 is cross connect of E1#1(Li0)b
  result = OtxDrvCreatePipe(64, &hPipe);
  DemoCheckRc(result);

  nTs = 17;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hPhysDevsGimle[iBrd].hLi[0],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);
  
  nTs = 2;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor[iBrd].hLi[0],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);
  
  DemoLogf("Pipe for TS%d: \n", nTs); DemoPrintPipeConnections(hPipe);

  // TS 3 is cross connect of E1#2(Li1)c
  result = OtxDrvCreatePipe(64, &hPipe);
  DemoCheckRc(result);

  nTs = 1;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hPhysDevsGimle[iBrd].hLi[1],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);
  
  nTs = 3;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor[iBrd].hLi[0],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);
  
  DemoLogf("Pipe for TS%d: \n", nTs); DemoPrintPipeConnections(hPipe);

  // TS 4 is cross connect of E1#2(Li1)d
  result = OtxDrvCreatePipe(64, &hPipe);
  DemoCheckRc(result);

  nTs = 16;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hPhysDevsGimle[iBrd].hLi[1],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);
  
  nTs = 4;
  memset(sqTimeSlotMask, 0x00, 32);
  sqTimeSlotMask[nTs] = 0xFF;
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor[iBrd].hLi[0],  -1, sqTimeSlotMask, 32);
  DemoCheckRc(result);

  DemoLogf("Pipe for TS%d: \n", nTs); DemoPrintPipeConnections(hPipe);


}


void DemoConnectDevices(
/*******************************************************************************

  @func

    Open, and parse the connection file. Make the connections according to the 
    specifications in that file.

*******************************************************************************/
  void
)
{
  OTX_INT32 nTs;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_HANDLE hPipe;
  FILE *fin;
  char szLine[256];
  char *pszToken;
  OTX_INT32 iGimleBrd;
  OTX_INT32 iGimleLi;
  OTX_INT32 iGimleTs;
  OTX_INT32 iThorBrd;
  OTX_INT32 iThorLi;
  OTX_INT32 iThorTs;
  OTX_INT32 nLineNo = 0;
  char *szWhiteSpace = " \t:,;\n";

  fin = fopen(g_demoEnv.conFileName, "rt");
  if (!fin) {
    DemoLogf("Error: Could not open timeslots connection file (\"%s\")\n", g_demoEnv.conFileName);
    DemoExitQuestion();
  }

  while (!feof(fin)) {
    nLineNo++;
    fgets(szLine, 256, fin);

    if (feof(fin)) break;

    pszToken = strtok(szLine, szWhiteSpace); if (!pszToken) continue; // Skip blank lines

    if (*pszToken == '#') {
      // comment, skip this line
      continue;
    }

    iGimleBrd = strtol(pszToken, NULL, 10);
    pszToken = strtok(NULL, szWhiteSpace); if (!pszToken) { DemoLogf("Syntax error on line #%d\n", nLineNo); DemoExitQuestion(); break; }
    iGimleLi = strtol(pszToken, NULL, 10);
    pszToken = strtok(NULL, szWhiteSpace); if (!pszToken) { DemoLogf("Syntax error on line #%d\n", nLineNo); DemoExitQuestion(); break; }
    iGimleTs = strtol(pszToken, NULL, 10);
    pszToken = strtok(NULL, szWhiteSpace); if (!pszToken) { DemoLogf("Syntax error on line #%d\n", nLineNo); DemoExitQuestion(); break; }
    iThorBrd = strtol(pszToken, NULL, 10);
    pszToken = strtok(NULL, szWhiteSpace); if (!pszToken) { DemoLogf("Syntax error on line #%d\n", nLineNo); DemoExitQuestion(); break; }
    iThorLi = strtol(pszToken, NULL, 10);
    pszToken = strtok(NULL, szWhiteSpace); if (!pszToken) { DemoLogf("Syntax error on line #%d\n", nLineNo); DemoExitQuestion(); break; }
    iThorTs = strtol(pszToken, NULL, 10);

    DemoLogf("Connecting Gimle#%d:Li#%d:TS#%d  to  Thor#%d:Li#%d:TS#%d\n", iGimleBrd, iGimleLi, iGimleTs, iThorBrd, iThorLi, iThorTs);

    result = OtxDrvCreatePipe(64, &hPipe);
    DemoCheckRc(result);
    
    nTs = iGimleTs;
    memset(sqTimeSlotMask, 0x00, 32);
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hPhysDevsGimle[iGimleBrd].hLi[iGimleLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(result);
    
    nTs = iThorTs;
    memset(sqTimeSlotMask, 0x00, 32);
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor[iThorBrd].hLi[iThorLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(result);
    
    DemoPrintPipeConnections(hPipe);

  }
    
  fclose(fin);


}





/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("GimleGroomer, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("E1/T1 Grooming Example program.\n");
  DemoLogf("For Gimle/Thor combos connected over H.100\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);


  {
    char szBuf[64];
    
    DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
    DemoLogf("Connection file is  %s\n", g_demoEnv.conFileName);

    DemoLogf("Gimle Board#%d is the clock source for the system\n", g_demoEnv.nGimleClkSourceBrd);
    
    switch (g_demoEnv.eGimleClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szBuf, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
    case OTX_CLOCK_SOURCE_LOCAL_8:
    case OTX_CLOCK_SOURCE_LOCAL_9:
    case OTX_CLOCK_SOURCE_LOCAL_10:
    case OTX_CLOCK_SOURCE_LOCAL_11:
    case OTX_CLOCK_SOURCE_LOCAL_12:
    case OTX_CLOCK_SOURCE_LOCAL_13:
    case OTX_CLOCK_SOURCE_LOCAL_14:
    case OTX_CLOCK_SOURCE_LOCAL_15:
      sprintf(szBuf, "Line Interface %d", (int)g_demoEnv.eGimleClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    default:
      strcpy(szBuf, "Unknown");
      break;
    }
    
    DemoLogf("Gimle Board#%d Clock Source is sourced from the %s source.\n", g_demoEnv.nGimleClkSourceBrd, szBuf);
    
    switch (g_demoEnv.eGimleT1E1Mode) {
    case OTX_T1E1_MODE_E1:
      strcpy(szBuf, "E1");
      break;
    case OTX_T1E1_MODE_T1:
      strcpy(szBuf, "T1");
      break;
    default:
      DemoLogf("Error: Unexpected environment value (g_demoEnv.eGimleT1E1Mode=%d)\n", g_demoEnv.eGimleT1E1Mode);
      DemoExitQuestion();
    }
    DemoLogf("Gimle Boards operate in %s %s mode\n", szBuf, g_demoEnv.bGimleReceiveEqualizer ? "Long-haul" : "Short-haul");
    switch (g_demoEnv.eGimleTermMode) {
    case OTX_T1E1_TERM_100_OHM:
      strcpy(szBuf, "100 ohm");
      break;
    case OTX_T1E1_TERM_75_OHM:
      strcpy(szBuf, "75 ohm");
      break;
    case OTX_T1E1_TERM_HIGHZ:
      strcpy(szBuf, "high");
      break;
    case OTX_T1E1_TERM_75_OHM_BALANCED:
      strcpy(szBuf, "75 ohm balanced");
      break;
    case OTX_T1E1_TERM_300_OHM:
      strcpy(szBuf, "300 ohm");
      break;
    default:
      DemoLogf("Error: Unexpected environment value (g_demoEnv.eGimleTermMode=%d)\n", g_demoEnv.eGimleTermMode);
      DemoExitQuestion();
    }
    DemoLogf("Gimle Boards terminated with %s impedance\n", szBuf);
    DemoLogf("Gimle Boards -20dB monitor mode is %s\n", g_demoEnv.bGimleMonitorMode ? "ON" : "OFF");
    DemoLogf("Gimle Boards additional amplifier is %s\n", g_demoEnv.bGimleAmplifier ? "ON" : "OFF");
    DemoLogf("Gimle Boards RX Input Threshold is %d mV\n", g_demoEnv.nGimleRxThreshold);
    if (g_demoEnv.eGimleT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1
      DemoLogf("Gimle Boards Line Code is %s\n", g_demoEnv.eGimleT1LineCode == OTX_T1_LC_AMI ? "AMI" : "B8ZS");
      switch (g_demoEnv.eGimleT1FrameFormat) {
      case OTX_T1_FF_ESF:
        strcpy(szBuf, "ESF");
        break;
      case OTX_T1_FF_F4:
        strcpy(szBuf, "F4");
        break;
      case OTX_T1_FF_F12:
        strcpy(szBuf, "SF (F12)");
        break;
      case OTX_T1_FF_F72:
        strcpy(szBuf, "SLC96 (F72)");
        break;
      default:
        DemoLogf("Error: Unexpected environment value (g_demoEnv.eGimleT1FrameFormat=%d)\n", g_demoEnv.eGimleT1FrameFormat);
        DemoExitQuestion();
      }
      DemoLogf("Gimle Boards Frame Format is %s\n", szBuf);
    } else {
      // E1
      DemoLogf("Gimle Boards Line Code is %s\n", g_demoEnv.eGimleE1LineCode == OTX_E1_LC_AMI ? "AMI" : "HDB3");
      switch (g_demoEnv.eGimleE1FrameFormat) {
      case OTX_E1_FF_DOUBLE_FRAME:
        strcpy(szBuf, "Double-Frame");
        break;
      case OTX_E1_FF_CRC4_MULTIFRAME:
        strcpy(szBuf, "CRC-4 Multiframe");
        break;
      case OTX_E1_FF_CRC4_MULTIFRAME_G706:
        strcpy(szBuf, "CRC-4 Multiframe (G.706)");
        break;
      default:
        DemoLogf("Error: Unexpected environment value (g_demoEnv.eGimleE1FrameFormat=%d)\n", g_demoEnv.eGimleE1FrameFormat);
        DemoExitQuestion();
      }
      DemoLogf("Gimle Boards Frame Format is %s\n", szBuf);
    }
    
    
    
    // ****** Thor Parameters ******
    switch (g_demoEnv.eThorT1E1Mode) {
    case OTX_T1E1_MODE_E1:
      strcpy(szBuf, "E1");
      break;
    case OTX_T1E1_MODE_T1:
      strcpy(szBuf, "T1");
      break;
    default:
      DemoLogf("Error: Unexpected environment value (g_demoEnv.eThorT1E1Mode=%d)\n", g_demoEnv.eThorT1E1Mode);
      DemoExitQuestion();
    }
    DemoLogf("Thor Boards operate in %s %s mode\n", szBuf, g_demoEnv.bThorReceiveEqualizer ? "Long-haul" : "Short-haul");
    
    
    switch (g_demoEnv.eThorTermMode) {
    case OTX_T1E1_TERM_100_OHM:
      strcpy(szBuf, "100 ohm");
      break;
    case OTX_T1E1_TERM_75_OHM:
      strcpy(szBuf, "75 ohm");
      break;
    case OTX_T1E1_TERM_HIGHZ:
      strcpy(szBuf, "high");
      break;
    case OTX_T1E1_TERM_75_OHM_BALANCED:
      strcpy(szBuf, "75 ohm balanced");
      break;
    case OTX_T1E1_TERM_300_OHM:
      strcpy(szBuf, "300 ohm");
      break;
    default:
      DemoLogf("Error: Unexpected environment value (g_demoEnv.eThorTermMode=%d)\n", g_demoEnv.eThorTermMode);
      DemoExitQuestion();
    }
    DemoLogf("Thor Boards terminated with %s impedance\n", szBuf);
    DemoLogf("Thor Boards -20dB monitor mode is %s\n", g_demoEnv.bThorMonitorMode ? "ON" : "OFF");
    DemoLogf("Thor Boards additional amplifier is %s\n", g_demoEnv.bThorAmplifier ? "ON" : "OFF");
    DemoLogf("Thor Boards RX Input Threshold is %d mV\n", g_demoEnv.nThorRxThreshold);
    if (g_demoEnv.eThorT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1
      DemoLogf("Thor Boards Line Code is %s\n", g_demoEnv.eThorT1LineCode == OTX_T1_LC_AMI ? "AMI" : "B8ZS");
      switch (g_demoEnv.eThorT1FrameFormat) {
      case OTX_T1_FF_ESF:
        strcpy(szBuf, "ESF");
        break;
      case OTX_T1_FF_F4:
        strcpy(szBuf, "F4");
        break;
      case OTX_T1_FF_F12:
        strcpy(szBuf, "SF (F12)");
        break;
      case OTX_T1_FF_F72:
        strcpy(szBuf, "SLC96 (F72)");
        break;
      default:
        DemoLogf("Error: Unexpected environment value (g_demoEnv.eThorT1FrameFormat=%d)\n", g_demoEnv.eThorT1FrameFormat);
        DemoExitQuestion();
      }
      DemoLogf("Thor Boards Frame Format is %s\n", szBuf);
    } else {
      // E1
      DemoLogf("Thor Boards Line Code is %s\n", g_demoEnv.eThorE1LineCode == OTX_E1_LC_AMI ? "AMI" : "HDB3");
      switch (g_demoEnv.eThorE1FrameFormat) {
      case OTX_E1_FF_DOUBLE_FRAME:
        strcpy(szBuf, "Double-Frame");
        break;
      case OTX_E1_FF_CRC4_MULTIFRAME:
        strcpy(szBuf, "CRC-4 Multiframe");
        break;
      case OTX_E1_FF_CRC4_MULTIFRAME_G706:
        strcpy(szBuf, "CRC-4 Multiframe (G.706)");
        break;
      default:
        DemoLogf("Error: Unexpected environment value (g_demoEnv.eThorE1FrameFormat=%d)\n", g_demoEnv.eThorE1FrameFormat);
        DemoExitQuestion();
      }
      DemoLogf("Thor Boards Frame Format is %s\n", szBuf);
    }
    
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver->Application communication
  DemoCreateEventQueue();

  // Connect to the Physical CT Bus
  DemoLogf("Open OTX_DEVICE_CT_BUS\n");
  result = OtxDrvOpenPhysicalDevice(0, OTX_DEVICE_CT_BUS, 0, 0, g_hEventQueue, &g_hCtBus); DemoCheckRc(result);

  // Enumerate, open, and configure all Gimle and Thor boards in the system
  DemoEnumerateDevices();

  // Set clocks source for all boards (Gimle and Thor)
  DemoSetClockSource();

#if 0
  DemoConnectDevicesHardcoded();
#else
  DemoConnectDevices();
#endif

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


