/* $Workfile: Dsp55Demo1.c $       $Revision: 1 $      $Modtime: 3/07/12 2:25p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    Dsp55Demo1.c

    A simple example console application demonstrating the use of the DSP SDK
    example program Dsp55Demo1.out. 
    
    The applications works with the following board combinations:

    OTX PCI Plus series boards with Vidar-55x4-ASM,

      Thor-2(4,8)-PCI-Plus + Vidar-55x4-ASM,

    or Thor-2-PCI-Express + Vidar-55x4-ASM,

    or Thor-2-PCIe + Vidar-55x4-ASM,

    or Thor-2-PCI-Express + Vidar-55x4-ASM,

    or Sleipnir-1-PCI.

    The example program opens DSP devices and loads the user written DSP SDK
    Example program "Dsp55Demo1.out" to these DSP devices.

    The program utilizes the I/O Control interface to control the blink
    frequency of the C5510 heart-beat LED. Simple keystroke commands are
    used to vary the blink frequency.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxDsp.h"

#include <Dsp55Demo1Io.h>   // IO Control codes for the DSP to host communication
#include <Dsp55Ioctl.h>  // IO Control codes for the DSP to host communication

#define DEMO_TASK_REF_DEBUG    501    // Task Reference for a non-blocking call
#define DEMO_DSP_MAX             4    // Max Number of DSP supported
#define DEMO_DEFAULT_LOG_FILE   "Dsp55Demo1.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_FAST_BLINK_COUNT_FACTOR 20
#define DEMO_SLOW_BLINK_COUNT_FACTOR 200

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL       bExitOnError;
  char           *logFileName;
  enum OtxPhysicalDeviceTypeE nBoardType;
  OTX_INT32      nDsps;
  OTX_INT32      nFirstDsp;
  OTX_BOOL       bAsmPresent;
  OTX_BOOL       bUseDspDebugger;
  int            boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoExitQuestion(void);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

static OsaMutexS g_HandleEventCritSec;

struct PhysDevs {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hAsmController0;
  OTX_HANDLE hDsp[DEMO_DSP_MAX];
} g_hPhysDevs;

struct LogicDevs {
  OTX_HANDLE hUserApp[DEMO_DSP_MAX];
} g_hLogicDevs;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Open the Physical Devices on the board
  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening OTX_DEVICE_THOR_2_PCI_PLUS board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening OTX_DEVICE_THOR_2_PCI_EXPRESS board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening OTX_DEVICE_THOR_4_PCI_PLUS board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening OTX_DEVICE_THOR_8_PCI_PLUS board\n");
      break;
    case OTX_DEVICE_THOR_8_PCIE:
      DemoLogf("Opening OTX_DEVICE_THOR_8_PCIe board\n");
      break;
    case OTX_DEVICE_SLEIPNIR_1_PCI:
      DemoLogf("Opening OTX_DEVICE_SLEIPNIR_1_PCI board\n");
      break;
    case OTX_DEVICE_HUGIN_2_PC104:
      DemoLogf("Opening OTX_DEVICE_HUGIN_2_PC104 board\n");
      break;
    case OTX_DEVICE_THOR_2_EXPRESSCARD:
      DemoLogf("Opening OTX_DEVICE_THOR_2_EXPRESSCARD board\n");
      break;
    case OTX_DEVICE_THOR_4_EXPRESSCARD:
      DemoLogf("Opening OTX_DEVICE_THOR_4_EXPRESSCARD board\n");
      break;
    default:
      DemoLogf("Opening '%d' board\n", g_demoEnv.nBoardType);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  if (g_demoEnv.nFirstDsp) {
    DemoLogf("DSP#%d will be referred to as DSP#0\n", g_demoEnv.nFirstDsp);
  }
  

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCIE:
    // Get a handle to to the ASM Daughter Board
    DemoLogf("Opening OTX_DEVICE_VIDAR_55x4_ASM board\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_VIDAR_55x4_ASM, 0, 0, g_hEventQueue, &g_hPhysDevs.hAsmController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hAsmController0, OTX_DEVICE_DSP, i+g_demoEnv.nFirstDsp, i+g_demoEnv.nFirstDsp, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
    }
    break;
  
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i+g_demoEnv.nFirstDsp, i+g_demoEnv.nFirstDsp, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
    }
    break;
  
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }


#if 0 // Overriding the Default HPI Buffer configuration (between the driver and the DSP)

  {
    OTX_INT32 nDsp;

    OTX_UINT32 nEventBufStartAddress      = 0x1010;
    OTX_UINT32 nEventBufSize              = 0x200;
    OTX_UINT32 nEventBufProcessingAddress = 0x100F;
    OTX_UINT32 nIoctlStartAddress         = 0x1300;
    OTX_UINT32 nIoctlSize                 = 0x200;
    OTX_BOOL bEventBufIncludeHandle     = OTX_FALSE;
    OTX_BOOL bEventBufIncludeTaskRef    = OTX_FALSE;
    OTX_BOOL bEventBufIncludeTimeStamp  = OTX_FALSE;
    OTX_BOOL bIoctlIncludeHandle        = OTX_TRUE;
    OTX_BOOL bIoctlIncludeTaskRef       = OTX_TRUE;

    for (nDsp = 0; nDsp < g_demoEnv.nDsps; nDsp++) {
      DemoLogf("Default Notification Event Buffer and Io Control attributes settings for DSP#%d:\n", nDsp);
      
      DemoPrintDspAttributes(nDsp);
      
      DemoLogf("Setting new attributes for DSP#%d\n", nDsp);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_ADDR, nEventBufStartAddress); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, nEventBufSize); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_PROCESSING_ADDR, nEventBufProcessingAddress); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_CODE_ADDR, nIoctlStartAddress); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_DATA_MAX_SIZE, nIoctlSize); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_HANDLE, bEventBufIncludeHandle); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TASKREF, bEventBufIncludeTaskRef); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TIMESTAMP, bEventBufIncludeTimeStamp); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_HANDLE, bIoctlIncludeHandle); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_TASKREF, bIoctlIncludeTaskRef); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("New Notification Event Buffer and Io Control attributes settings for DSP#%d:\n", nDsp);
      
      DemoPrintDspAttributes(nDsp);
    }

  }

#endif

  {
    // Load DSP Program
    char *pszDspProgram = "Dsp55Demo1.out";
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Loading program %s to DSP#%d\n", pszDspProgram, i);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], pszDspProgram, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    
    if (g_demoEnv.bUseDspDebugger) {
      DemoLogf("Please load and run DSP program %s to DSP#%d-%d in your DSP Debugger\n", pszDspProgram, i, g_demoEnv.nDsps-1);
      DemoLogf("Press any key when all the DSP program are running...\n");
      OsaConsoleGetCh();

      // At this point we need to tell the host to restart the programs 
      // (to issue the initial IO commands to the DSP again)
      for (i = 0; i < g_demoEnv.nDsps; i++) {
        DemoLogf("Restarting program on DSP#%d\n", i);
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }
      // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
    }

  }
}


void DemoPrintDspAttributes(OTX_INT32 nDsp)
/*******************************************************************************
  @func
    Queries and prints some of the DSP attributes
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_UINT32 nAttribValueUint32;
  OTX_BOOL nAttribValueBool;
  DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_ADDR, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf Start Address : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Size (in words)         : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_PROCESSING_ADDR, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Processing Address      : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_CODE_ADDR, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  IoCtl Start Address     : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_DATA_MAX_SIZE, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Size (in words)         : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_HANDLE, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf Handle        : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TASKREF, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf TaskRef       : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TIMESTAMP, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf TimeStamp     : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_HANDLE, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  IoCtl Handle            : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_TASKREF, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  IoCtl TaskRef           : %s\n", nAttribValueBool ? "Yes" : "No");
  
}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  if (g_hPhysDevs.hAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
  }

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}


void DemoCreateLogicalDevices(void)
/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    // Create one User Application Logical Devices per DSP
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, i, &g_hLogicDevs.hUserApp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hUserApp[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
  }
}




static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampUs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  // Calculate number of microseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampUs.nUint64 = nTimeStamp.nUint64 / 10 - (nTimeStampSec.nUint64 * 1000000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
#if 0    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
#else
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampUs.sUint32.nLow
      );
#endif
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

void *DemoIdleFunction(
/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  OsaMutexInit(&g_HandleEventCritSec);

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  OsaMutexLock(&g_HandleEventCritSec);

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the Dsp55Demo1.out logical devices
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8192];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8192, &nBytesReturned, &nIoControlCode);
      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        switch(nIoControlCode) {
        case DEMO1_NOTIFY_DEBUG_DATA:
          {
            Demo1NotifyGetDebugDataS *pDataS = (Demo1NotifyGetDebugDataS *)buf;
            if (nBytesReturned != sizeof(Demo1NotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = %.4X\n", pDataS->nData0);
              DemoLogf("nData1 = %.4X\n", pDataS->nData1);
              DemoLogf("nData2 = %.4X\n", pDataS->nData2);

              DemoLogf("sqBuf[299] = %.4X\n", pDataS->sqBuf[299]);
            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }

      }

    } else {

      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }

  OsaMutexUnlock(&g_HandleEventCritSec);


}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  static OTX_INT32      nCurrentUserApp = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-3 Select active DSP User Application\n");
    DemoLogf(" b   Set Blink Count Factor to %d for current DSP User Application \n", DEMO_FAST_BLINK_COUNT_FACTOR);
    DemoLogf(" B   Set Blink Count Factor to %d for current DSP User Application \n", DEMO_SLOW_BLINK_COUNT_FACTOR);
    DemoLogf(" n   Set Count Of Debug Notifications to be sent when 'd' is pressed (default is 1)\n");
    DemoLogf(" e   Enable DSP User Application (Start blinking) on current Dsp\n");
    DemoLogf(" E   Disable DSP User Application (Stop blinking) on current Dsp\n");
    DemoLogf(" d   Get Debug data from the current DSP User Application\n");
    DemoLogf(" D   Get HPI Event Buf Diagnostics data from the current DSP User Application\n");
    DemoLogf(" f   Get Blink Count from the current DSP User Application\n");
    DemoLogf(" r   Reset Blink Count to 0 in current DSP User Application\n");
    DemoLogf(" x   Perform a series of OtxDspIoControl() calls\n");
    DemoLogf(" p   Dump the entire DSP->Host Event buffer for the current DSP\n");
    DemoLogf(" P   Dump the entire Host->DSP IoControl buffer for the current DSP\n");
    DemoLogf(" i   Get DSP Chip information (Revision)\n");
    break;

  case 'b':
    {
      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_FAST_BLINK_COUNT_FACTOR, nCurrentUserApp);

      sIn.nBlinkCountFactor = DEMO_FAST_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 6;
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'x':
    {
      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      static OTX_INT32 nLoopTimes = 1;
      OTX_INT32 i;
      char strBuf[12];
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_FAST_BLINK_COUNT_FACTOR, nCurrentUserApp);
      DemoLogf("How many times? (%d)", nLoopTimes);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nLoopTimes = strtol(strBuf, NULL, 0);
      }

      sIn.nBlinkCountFactor = DEMO_FAST_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 6;

      DemoLogf("Performing %d OtxDspIoControl() calls \n ", nLoopTimes);
      for (i = 0; i < nLoopTimes; i++) {
        result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      }

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'B':
    {
      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_SLOW_BLINK_COUNT_FACTOR, nCurrentUserApp);

      sIn.nBlinkCountFactor = DEMO_SLOW_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 7;
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'n':
    {
      Demo1IoInSetCoDebugNotificationsS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      char szBuf[12];

      DemoLogf("Enter number of Debug Notifications to be sent when 'd' is pressed: ");
      OsaConsoleGetStr(szBuf);
      sIn.coDebugDataNotifications = (unsigned short)strtol(szBuf, NULL, 0);

      DemoLogf("Sending coDebugDataNotifications(%d) to User App #%d ...", sIn.coDebugDataNotifications, nCurrentUserApp);
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_CO_DEBUG_NOTIFICATIONS, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'e':
    {
      DemoLogf("Enable Blinking for User App #%d ...", nCurrentUserApp);
      result = OtxDrvEnable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'E':
    {
      DemoLogf("Disable Blinking for User App #%d ...", nCurrentUserApp);
      result = OtxDrvDisable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      DemoLogf("Reset for User App #%d ...", nCurrentUserApp);
      result = OtxDrvReset(g_hLogicDevs.hUserApp[nCurrentUserApp]); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'd':
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for User App #%d (non-blocking call)...", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, DEMO_TASK_REF_DEBUG, 5000); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'f':
    {
      OTX_UINT32 nBytesReturned;
      Demo1IoOutGetBlinkFreqS dataS;

      DemoLogf("Get Blink Frequency for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_BLINK_FREQ, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo1IoOutGetBlinkFreqS)) {
          DemoLogf("Unexpected data length for DEMO1_IOCTL_GET_BLINK_FREQ (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo1 Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("BlinkCountMax = %d\n", (OTX_UINT32)(dataS.nBlinkCountMaxHigh << 16) + dataS.nBlinkCountMaxLow);
          DemoLogf("BlinkCount    = %d\n", (OTX_UINT32)(dataS.nBlinkCountHigh << 16) + dataS.nBlinkCountLow);
        }
      }


      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'i':
    {
      OTX_UINT32 nBytesReturned;
      Demo1IoOutGetChipInfoS dataS;

      DemoLogf("Get DSP Chip Information (Revision) for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_CHIP_INFO, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo1IoOutGetChipInfoS)) {
          DemoLogf("Unexpected data length for DEMO1_IOCTL_GET_CHIP_INFO (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo1 Error Code: %d\n", dataS.nResult);
        } else {
          switch (dataS.nRevId) {
          case 0x8050:
            DemoLogf("Silicon Version 1.x (0x%x)\n", dataS.nRevId);
            break;
          case 0x6510:
            DemoLogf("Silicon Version 2.0 (0x%x)\n", dataS.nRevId);
            break;
          case 0x6511:
            DemoLogf("Silicon Version 2.1 (0x%x)\n", dataS.nRevId);
            break;
          case 0x6512:
            DemoLogf("Silicon Version 2.2 (0x%x)\n", dataS.nRevId);
            break;
          default:
            DemoLogf("Silicon Version <unknown> (0x%x)\n", dataS.nRevId);
            break;
          }   
        }
      }


      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'p':
    {
      OTX_UINT32 i;
      OTX_UINT16 sqReadBuf[OTXDSP_SDK_EVENT_BUF_LEN];

      DemoLogf("Dump the entire DSP->Host Event buffer for Dsp #%d \n", nCurrentUserApp);

      result = OtxDspReadHpi(g_hPhysDevs.hDsp[nCurrentUserApp], OTXDSP_SDK_EVENT_BUF_ADDR, OTXDSP_SDK_EVENT_BUF_LEN, sqReadBuf); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      for (i = 0; i < OTXDSP_SDK_EVENT_BUF_LEN; i++) {
        if ((i % 8) == 0) DemoLogf("\n%.8X: ", OTXDSP_SDK_EVENT_BUF_ADDR+i);
        DemoLogf("%.4X ", sqReadBuf[i]);
      }
      DemoLogf("\nDone.\n");
    }
    break;

  case 'P':
    {
      OTX_UINT32 i;
      OTX_UINT16 sqReadBuf[OTXDSP_SDK_IOCTL_DATA_MAX_SIZE+8];

      DemoLogf("Dump the entire Host->DSP IoControl buffer for Dsp #%d \n", nCurrentUserApp);

      result = OtxDspReadHpi(g_hPhysDevs.hDsp[nCurrentUserApp], OTXDSP_SDK_IOCTL_CODE_ADDR, OTXDSP_SDK_IOCTL_DATA_MAX_SIZE+8, sqReadBuf); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      for (i = 0; i < OTXDSP_SDK_IOCTL_DATA_MAX_SIZE+8; i++) {
        if ((i % 8) == 0) DemoLogf("\n%.8X: ", OTXDSP_SDK_IOCTL_CODE_ADDR+i);
        DemoLogf("%.4X ", sqReadBuf[i]);
      }
      DemoLogf("\nDone.\n");
    }
    break;

  case 'D':
    {
      OTX_UINT32 nBytesReturned;
      Demo1IoOutGetHpiEventBufDiagS dataS;

      DemoLogf("Get HPI Event Buf Diagnostics for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_HPI_EVENT_BUF_DIAG, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo1IoOutGetHpiEventBufDiagS)) {
          DemoLogf("Unexpected data length for DEMO1_IOCTL_GET_HPI_EVENT_BUF_DIAG (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo1 Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("cwFreeBuffer    = %d\n", dataS.cwFreeBuffer);
          DemoLogf("cwMinFreeBuffer = %d\n", dataS.cwMinFreeBuffer);
          DemoLogf("coLostEvents    = %d\n", dataS.coLostEvents);
        }
      }


      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
    // Select current DSP
    nCurrentUserApp = commandChar - '0';
    DemoLogf("DSP User Application #%d Selected ....\n", nCurrentUserApp);
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command (%d). Press '?' for help, 'q' to quit\n", commandChar);
    break;
  }

  return 0;
}



static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *pEnv)
/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
{
  short  i;

  // Set reasonable defaults before reading from command line
  pEnv->pszInputStr     = NULL;
  pEnv->bExitOnError    = OTX_FALSE;
  pEnv->logFileName     = DEMO_DEFAULT_LOG_FILE;
  pEnv->nBoardType      = OTX_DEVICE_SLEIPNIR_1_PCI;
  pEnv->nDsps           = 1; // for now
  pEnv->nFirstDsp       = 0;
  pEnv->boardNo         = DEMO_DEFAULT_BOARD_NO;
  pEnv->bUseDspDebugger = OTX_FALSE;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      pEnv->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          pEnv->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          pEnv->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'N') {   // Board number
          pEnv->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // Number of DSPs to use
          pEnv->nDsps = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'F') {   // DSP number of the first to use (in case we do not want to use DSP0 as DSP0).
          pEnv->nFirstDsp = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          pEnv->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            pEnv->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'P') {
            pEnv->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'W') {
            pEnv->nBoardType = OTX_DEVICE_THOR_8_PCIE;
          } else if ( *(argv[i]+2) == 'S') {
            pEnv->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
          } else if ( *(argv[i]+2) == 'T') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'E') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if ( *(argv[i]+2) == 'H') {
            pEnv->nBoardType = OTX_DEVICE_HUGIN_2_PC104;
          } else if ( *(argv[i]+2) == 'C') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
          } else if ( *(argv[i]+2) == 'R') {
            pEnv->nBoardType = OTX_DEVICE_THOR_4_EXPRESSCARD;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /N<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[4|P|W|S|T|E|H|C|R]  : Board Type used for the demo, where\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board with Vidar-55x4-ASM\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board with Vidar-55x4-ASM\n");
  DemoLogf("                  W = Use Thor-8-PCIe Board with Vidar-55x4-ASM\n");
  DemoLogf("                  S = Use Sleipnir-1-PCI Board with 5510 DSP\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board with Vidar-55x4-ASM\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board with Vidar-55x4-ASM\n");
  DemoLogf("                  H = Use Hugin-2-PC104 Board\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                  R = Use Thor-4-ExpressCard Board\n");
  DemoLogf(" /D             : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /X<number>     : Number of DSPs to configure (1-4). Default is 1.\n");
  DemoLogf(" /F<number>     : DSP number of the first DSP to configure. Default is 0.\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);
}



int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("Dsp55Demo1, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DSP C55x SDK Demo Program for OTX PCI and PCIe boards.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Connect to the Driver
  DemoLogf("Connecting lib... ");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Query driver revision
  {
    char szVersionInfo[255];
    OtxDrvGetAttributeValueSz(0, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened physical devices...\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened logical devices...\n");
  DemoEnableLogicalDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

#if 0 // Just for testing
  DemoLogf("Sleeping 2 seconds...\n");
  OsaThreadSleep(2000);
  DemoLogf("Exiting...\n");
  OtxDrvReleaseLib();
  OsaConsoleExit();
  return 0;
#endif

  DemoLogf("\nMain loop running. Press 'h' for help\n");
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

